/* ball.c
**
**    code for creating the GL-ball display-lists
**    Copyright (C) 2001  Florian Berger
**    Email: harpin_floh@yahoo.de, florian.berger@jk.uni-linz.ac.at
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License Version 2 as
**    published by the Free Software Foundation;
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/

#include <GL/glu.h>
#include <GL/freeglut.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>  /* open */
#include <sys/stat.h>   /* open */
#include <fcntl.h>      /* open */
#include <unistd.h>     /* close */
#include "globe.h"
#include "png_loader.h"
#include "jpg_loader.h"
#include "options.h"
#include "markers_parser.h"


#undef DEBUG

#define USE_VERTEX_ARRAYS 1


#define GRID_R_FACT 1.00000
//#define GRID_R_FACT 0.99

static GLfloat col_spec[4] = {0.0, 0.0, 0.0, 1.0}; /* dont need any specular because of reflections */
//static GLfloat col_refl[4] = {1.0, 1.0, 1.0, 0.2};
static GLfloat col_diff[4] = {1.7, 1.7, 1.7, 1.0};
static GLfloat col_amb [4] = {0.4, 0.4, 0.4, 1.0};
//static GLfloat col_amb [4] = {0.0, 0.0, 0.0, 1.0};
//static GLfloat col_shad[4] = {1.0, 0.0, 0.0, 0.5};
static char globe_map_file[256];


void normalize(float *v) /*FOLD00*/
{
    GLfloat d;
    d = sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
    v[0] /= d; v[1] /= d; v[2] /= d;
}


void rescale(float *v, float r)
{
    GLfloat d;
    d = sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2])/r;
    v[0] /= d; v[1] /= d; v[2] /= d;
}


/*
void normalize_my(myfloat *v)
{
    GLfloat d;
    d = sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
    v[0] /= d; v[1] /= d; v[2] /= d;
}
*/


static int in_array( ElemArray * array, GLfloat * v, GLfloat tex_x, GLfloat tex_y )
{
    int i;

    for(i=0;i<array->vnr;i++){
        if( v[0]==array->vert[i*3+0] &&
            v[1]==array->vert[i*3+1] &&
            v[2]==array->vert[i*3+2] &&
            tex_x==array->tex[i*2+0] &&
            tex_y==array->tex[i*2+1]
          ){
            return i;
        }
    }

    return -1;
}


void createvertex( float * v, float * n,
                   float tex_x, float tex_y, ElemArray * array )
{
    int pos;
    if ( array == NULL ){
        glTexCoord2f( tex_x, tex_y );
        glNormal3fv(n);
        glVertex3fv(v);
    } else {
        pos = in_array( array,v, tex_x, tex_y );
        if( pos!=-1 ){
            array->index[array->indnr] = pos;
            array->indnr++;
        } else {
            array->vert[array->vnr*3+0]=v[0];
            array->vert[array->vnr*3+1]=v[1];
            array->vert[array->vnr*3+2]=v[2];
            array->norm[array->vnr*3+0]=n[0];
            array->norm[array->vnr*3+1]=n[1];
            array->norm[array->vnr*3+2]=n[2];
            array->tex[array->vnr*2+0]=tex_x;
            array->tex[array->vnr*2+1]=tex_y;
            array->index[array->indnr] = array->vnr;
            array->vnr++;
            array->indnr++;
            fprintf(stderr,"createvertex: array->vnr=%d\n",array->vnr);
        }
    }
}

// because used in ball_subdivide_rec
void ball_subdivide9_rec(float *v1, float *v2, float *v3, const int depth, float r, ElemArray *array);

void ball_subdivide_rec(float *v1, float *v2, float *v3, const int depth, float r, ElemArray *array)
/* if array=0 just make vertices */
{
    GLfloat v12[3], v23[3], v31[3], v123[3];
    GLfloat n1[3], n2[3], n3[3];
    GLint i;
    int otherside;
    double xt,yt,rho;

//    fprintf(stderr,"depth=%d\n",depth);
    if (depth == 0) {

        for(i=0;i<3;i++){
            n1[i]=v1[i];   n2[i]=v2[i];   n3[i]=v3[i];
        }
        normalize(n1);  normalize(n2);  normalize(n3);
        rescale(v1,r);  rescale(v2,r);  rescale(v3,r);

        otherside = ( v1[2]>0.0 || v2[2]>0.0 || v3[2]>0.0 );
//        glEnable(GL_NORMALIZE);
//        glTexCoord2f(0.5+0.5*v1[0], 0.5-0.5*(otherside?v1[1]:-v1[1]) );
        rho = sqrt(v1[0]*v1[0]+v1[1]*v1[1]);
        xt=asin(rho)/M_PI*2.3*v1[0]/rho;
        yt=asin(rho)/M_PI*2.3*v1[1]/rho;
        xt=v1[0]/r;
        yt=v1[1]/r;
        createvertex( v1, n1, 0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt), array);
//        glTexCoord2f(0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt) );
//        glNormal3fv(n1);
//        glVertex3fv(v1);

        rho = sqrt(v2[0]*v2[0]+v2[1]*v2[1]);
        xt=asin(rho)/M_PI*2.3*v2[0]/rho;
        yt=asin(rho)/M_PI*2.3*v2[1]/rho;
        xt=v2[0]/r;
        yt=v2[1]/r;
        createvertex( v2, n2, 0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt), array);
//        glTexCoord2f(0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt) );
//        glNormal3fv(n2);
//        glVertex3fv(v2);

        rho = sqrt(v3[0]*v3[0]+v3[1]*v3[1]);
        xt=asin(rho)/M_PI*2.3*v3[0]/rho;
        yt=asin(rho)/M_PI*2.3*v3[1]/rho;
        xt=v3[0]/r;
        yt=v3[1]/r;
        createvertex( v3, n3, 0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt), array);
//        glTexCoord2f(0.5+0.5*xt, 0.5-0.5*(otherside?yt:-yt) );
//        glNormal3fv(n3);
//        glVertex3fv(v3);

    } else if( depth==1 ){
        for (i = 0; i < 3; i++) v123[i] = v1[i]+v2[i]+v3[i];
        rescale(v123,r);
        ball_subdivide_rec( v123, v1, v2, depth-1, r, array);
        ball_subdivide_rec( v123, v2, v3, depth-1, r, array);
        ball_subdivide_rec( v123, v3, v1, depth-1, r, array);
    } else if( depth==3 ){
        ball_subdivide9_rec( v1, v2, v3, depth-2, r, array);
    } else {
        for (i = 0; i < 3; i++) {
            v12[i] = v1[i]+v2[i];
            v23[i] = v2[i]+v3[i];
            v31[i] = v3[i]+v1[i];
        }
        rescale(v12,r);
        rescale(v23,r);
        rescale(v31,r);
        ball_subdivide_rec( v1, v12, v31, depth-2, r, array);
        ball_subdivide_rec( v2, v23, v12, depth-2, r, array);
        ball_subdivide_rec( v3, v31, v23, depth-2, r, array);
        ball_subdivide_rec(v12, v23, v31, depth-2, r, array);
    }
//    fprintf(stderr,"back again all (depth=%d)\n",depth);
}


void ball_subdivide9_rec(float *v1, float *v2, float *v3, const int depth, float r, ElemArray *array)
{
    GLfloat v112[3], v122[3], v223[3], v233[3], v331[3], v311[3], v123[3];
    GLint i;

//    fprintf(stderr,"depth=%d\n",depth);
    if (depth == 0) {
        ball_subdivide_rec(v1,v2,v3,0,r, array); // draw the triangle
    } else {
        for (i = 0; i < 3; i++) {
            v112[i] = v1[i]+v1[i]+v2[i];
            v122[i] = v1[i]+v2[i]+v2[i];
            v223[i] = v2[i]+v2[i]+v3[i];
            v233[i] = v2[i]+v3[i]+v3[i];
            v331[i] = v3[i]+v3[i]+v1[i];
            v311[i] = v3[i]+v1[i]+v1[i];
            v123[i] = v1[i]+v2[i]+v3[i];
        }
        rescale(v112,r);
        rescale(v122,r);
        rescale(v223,r);
        rescale(v233,r);
        rescale(v331,r);
        rescale(v311,r);
        rescale(v123,r);
        ball_subdivide9_rec( v1  , v112, v311, depth-1, r, array);
        ball_subdivide9_rec( v112, v122, v123, depth-1, r, array);
        ball_subdivide9_rec( v112, v123, v311, depth-1, r, array);
        ball_subdivide9_rec( v311, v123, v331, depth-1, r, array);
        ball_subdivide9_rec( v122, v2  , v223, depth-1, r, array);
        ball_subdivide9_rec( v122, v223, v123, depth-1, r, array);
        ball_subdivide9_rec( v123, v223, v233, depth-1, r, array);
        ball_subdivide9_rec( v123, v233, v331, depth-1, r, array);
        ball_subdivide9_rec( v331, v233, v3  , depth-1, r, array);
    }
//    fprintf(stderr,"back again all (depth=%d)\n",depth);
}



void create_ball_icosa( float r, int ddepth, int id ) /*FOLD00*/
{
#define X .525731112119133606
#define Z .850650808352039932

    int i;

    static GLfloat vdata[12][3] = {
        {-X, 0.0, Z}, {X, 0.0, Z}, {-X, 0.0, -Z}, {X, 0.0, -Z},
        {0.0, Z, X}, {0.0, Z, -X}, {0.0, -Z, X}, {0.0, -Z, -X},
        {Z, X, 0.0}, {-Z, X, 0.0}, {Z, -X, 0.0}, {-Z, -X, 0.0}
    };

    static GLint tindices[20][3] = {
        {0,4,1}, {0,9,4}, {9,5,4}, {4,5,8}, {4,8,1},
        {8,10,1}, {8,3,10}, {5,3,8}, {5,2,3}, {2,7,3},
        {7,10,3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1,6},
        {6,1,10}, {9,0,11}, {9,11,2}, {9,2,5}, {7,2,11}
    };


    glDisable( GL_NORMALIZE );   /* remove this */
    glNewList(id, GL_COMPILE);
//      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
//      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glPushMatrix();
//      glScalef( r, r, r );
//      glEnable( GL_NORMALIZE );   /* remove this */
      glBegin(GL_TRIANGLES);
         for (i = 0; i < 20; i++) {
             ball_subdivide_rec(
                                &vdata[tindices[i][0]][0],
                                &vdata[tindices[i][1]][0],
                                &vdata[tindices[i][2]][0],
                                ddepth, r, NULL
                               );
         }
      glEnd();
      glPopMatrix();
//      glDisable( GL_NORMALIZE );   /* remove this */
    glEndList();

}



void create_ball_icosa_array( float r, int ddepth, int id )
{
#define X .525731112119133606
#define Z .850650808352039932

    ElemArray * array;
    int i,pnr,fnr;

    static GLfloat vdata[12][3] = {
        {-X, 0.0, Z}, {X, 0.0, Z}, {-X, 0.0, -Z}, {X, 0.0, -Z},
        {0.0, Z, X}, {0.0, Z, -X}, {0.0, -Z, X}, {0.0, -Z, -X},
        {Z, X, 0.0}, {-Z, X, 0.0}, {Z, -X, 0.0}, {-Z, -X, 0.0}
    };

    static GLint tindices[20][3] = {
        {0,4,1}, {0,9,4}, {9,5,4}, {4,5,8}, {4,8,1},
        {8,10,1}, {8,3,10}, {5,3,8}, {5,2,3}, {2,7,3},
        {7,10,3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1,6},
        {6,1,10}, {9,0,11}, {9,11,2}, {9,2,5}, {7,2,11}
    };

    static int init = 0;


    if(!init){
        init=1;
         for (i = 0; i < 12; i++) {
             rescale(vdata[i],r);
         }
    }

    if(ddepth%2==0){
        fnr = 20*(1<<(ddepth/2*2));
        pnr = 10*((1<<(ddepth/2*2))-1) + 12;
    } else if(ddepth!=1) {
        fnr = 20*(1<<((ddepth-3)/2*2));
        pnr = 10*((1<<((ddepth-3)/2*2))-1) + 12;
        pnr += fnr * 4;
        fnr *= 9;
    } else if(ddepth==1) {
        fnr = 20*(1<<(ddepth/2*2));
        pnr = 10*((1<<(ddepth/2*2))-1) + 12;
        pnr += fnr; fnr *= 3;
    }
    pnr+=(int)sqrt(pnr)*2;  /* could get narrow */

    array = (ElemArray *)malloc( sizeof(ElemArray) );
    array->vert  = (GLfloat *)malloc( pnr*3*sizeof(GLfloat) );
    array->norm  = (GLfloat *)malloc( pnr*3*sizeof(GLfloat) );
    array->tex   = (GLfloat *)malloc( pnr*2*sizeof(GLfloat) );
    array->index = (int *)    malloc( fnr*3*sizeof(int)     );

    array->indnr=0;
    array->vnr=0;

    fprintf(stderr,"fnr*3=%d\n",fnr*3);
    fprintf(stderr,"pnr=%d\n",pnr);

    for (i = 0; i < 20; i++) {
        ball_subdivide_rec(
                           &vdata[tindices[i][0]][0],
                           &vdata[tindices[i][1]][0],
                           &vdata[tindices[i][2]][0],
                           ddepth, r, array
                          );
    }

    fprintf(stderr,"array->indnr=%d\n",array->indnr);
    fprintf(stderr,"array->vnr=%d\n",array->vnr);
/*
    glNewList(id, GL_COMPILE);
      glPushMatrix();
      glBegin(GL_TRIANGLES);
         for (i = 0; i < array->indnr; i++) {
             glTexCoord2fv( &array->tex [array->index[i]*2] );
             glNormal3fv(   &array->norm[array->index[i]*3] );
             glVertex3fv(   &array->vert[array->index[i]*3] );
             fprintf(stderr,"v=<%f,%f,%f>\n",
                     array->vert[array->index[i]*3],
                     array->vert[array->index[i]*3+1],
                     array->vert[array->index[i]*3+2] );
         }
      glEnd();
      glPopMatrix();
    glEndList();
*/

#if 1
    glNewList(id, GL_COMPILE);
      glPushMatrix();
        /* call array */
        glEnableClientState( GL_VERTEX_ARRAY );
        glEnableClientState( GL_NORMAL_ARRAY );
        glEnableClientState( GL_TEXTURE_COORD_ARRAY );
        glVertexPointer( 3, GL_FLOAT, 0, array->vert );
        glNormalPointer( GL_FLOAT, 0, array->norm );
        glTexCoordPointer( 2, GL_FLOAT, 0, array->tex );

        fprintf(stderr,"indnr=%d\n",array->indnr);
        fprintf(stderr,"vnr=%d\n",array->vnr);

        glDrawElements( GL_TRIANGLES, array->indnr, GL_UNSIGNED_INT, array->index );

//      glEnd();
      glPopMatrix();
    glEndList();
#endif
    
    fprintf(stderr,"array compiled %d, NE=%d\n",glGetError(),GL_NO_ERROR);
}



int make_globe_old( double r, int nth, int nph, int dx, int dy, char * filename )
/* returns id */
{
    int ith,iph,x,y;
    char * data;
    float th1,th2,ph1,ph2;
    int globe_id;
    int allw,allh,alld;
    char * alldata;
    static unsigned int texbind[10000];

    load_png( filename, &allw, &allh, &alld, &alldata );

    data = (char *)malloc(3*dx*dy);

    glGenTextures(nth*nph, texbind);

    for(ith=0;ith<nth;ith++){
        for(iph=0;iph<nph;iph++){
            /* get new quad texture */
            for(y=0;y<dy;y++){
                for(x=0;x<dx;x++){
                    data[(y*dx+x)*3+0]=alldata[((ith*dy+y)*allw+iph*dx+x)*3+0];
                    data[(y*dx+x)*3+1]=alldata[((ith*dy+y)*allw+iph*dx+x)*3+1];
                    data[(y*dx+x)*3+2]=alldata[((ith*dy+y)*allw+iph*dx+x)*3+2];
//                    data[(y*dx+x)*3+0]=rand()&255;
//                    data[(y*dx+x)*3+1]=rand()&255;
//                    data[(y*dx+x)*3+2]=rand()&255;
                }
            }
            glBindTexture(GL_TEXTURE_2D,texbind[ith*nph+iph]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, options_tex_min_filter);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, options_tex_mag_filter);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
            gluBuild2DMipmaps(GL_TEXTURE_2D, 3, dx, dy, GL_RGB,
                              GL_UNSIGNED_BYTE, data);
        }
    }

    globe_id = glGenLists(1);
    glNewList(globe_id, GL_COMPILE);

//    glBegin(GL_QUADS);

    for(ith=0;ith<nth;ith++){
        th1 = M_PI/(double)nth*(double)ith;
        th2 = M_PI/(double)nth*(double)(ith+1);
        for(iph=0;iph<nph;iph++){
            ph1 = 2.0*M_PI/(double)nph*(double)iph;
            ph2 = 2.0*M_PI/(double)nph*(double)(iph+1);

            glBindTexture(GL_TEXTURE_2D, texbind[ith*nph+iph]);
            glBegin(GL_QUADS);
            /* draw quad */
            glTexCoord2f( 0.0,0.0 );
            glNormal3f(   sin(th1)*cos(ph1),   sin(th1)*sin(ph1),   cos(th1) );
            glVertex3f( r*sin(th1)*cos(ph1), r*sin(th1)*sin(ph1), r*cos(th1) );
            glTexCoord2f( 1.0,0.0 );
            glNormal3f(   sin(th1)*cos(ph2),   sin(th1)*sin(ph2),   cos(th1) );
            glVertex3f( r*sin(th1)*cos(ph2), r*sin(th1)*sin(ph2), r*cos(th1) );
            glTexCoord2f( 1.0,1.0 );
            glNormal3f(   sin(th2)*cos(ph2),   sin(th2)*sin(ph2),   cos(th2) );
            glVertex3f( r*sin(th2)*cos(ph2), r*sin(th2)*sin(ph2), r*cos(th2) );
            glTexCoord2f( 0.0,1.0 );
            glNormal3f(   sin(th2)*cos(ph1),   sin(th2)*sin(ph1),   cos(th2) );
            glVertex3f( r*sin(th2)*cos(ph1), r*sin(th2)*sin(ph1), r*cos(th2) );
            glEnd();

        }
    }

    glEndList();

    free(data);
    free(alldata);

    return(globe_id);
}


int make_globe_old2( double r, int nth, int nph, int dx, int dy, char * filename )
/* returns id */
{
    int ith,iph,x,y,w,h,dth,dph,qthnr,qphnr,texnr,qth,qph;
    unsigned char * data;
    float th1,th2,ph1,ph2, tex_x1,tex_x2,tex_y1,tex_y2;
    int globe_id;
    int allw,allh,alld;
    unsigned char * alldata;
    static unsigned int texbind[10000];

    if      ( strcasecmp(&filename[strlen(filename)-4],".jpg")==0 ){
        allw=1536;  allh=768;
        read_jpeg( filename, &allw, &allh, &alldata );
        fprintf(stderr,"loaded jpg-file w/h:%d/%d\n",allw,allh);
    }else if( strcasecmp(&filename[strlen(filename)-4],".png")==0 ){
        load_png( filename, &allw, &allh, &alld, (char **)&alldata );
    }
//    alld=3;

    data = (char *)malloc(3*dx*dy);

    dth=dy/(allh/nth);
    dph=dx/(allw/nph);

    texnr=-1;
    for(ith=0;ith<nth;ith+=dth){
        for(iph=0;iph<nph;iph+=dph){
            /* get new quad texture */
            w=dx; h=dy;
            if( (iph/dph)*dx+w > allw ) w = allw % dx;
            if( (ith/dth)*dy+h > allh ) h = allh % dy;
            for(y=0;y<h;y++){
                for(x=0;x<w;x++){
                    data[(y*w+x)*3+0]=alldata[(((ith/dth)*dy+y)*allw+(iph/dph)*dx+x)*3+0];
                    data[(y*w+x)*3+1]=alldata[(((ith/dth)*dy+y)*allw+(iph/dph)*dx+x)*3+1];
                    data[(y*w+x)*3+2]=alldata[(((ith/dth)*dy+y)*allw+(iph/dph)*dx+x)*3+2];
//                    data[(y*dx+x)*3+0]=rand()&255;
//                    data[(y*dx+x)*3+1]=rand()&255;
//                    data[(y*dx+x)*3+2]=rand()&255;
                }
            }
            texnr++;
            fprintf(stderr,"tex = %d,   w = %d,   h = %d\n",texnr,w,h);
            glGenTextures( 1 , &texbind[texnr] );
            glBindTexture(GL_TEXTURE_2D,texbind[texnr]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, options_tex_min_filter);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, options_tex_mag_filter);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
            gluBuild2DMipmaps(GL_TEXTURE_2D, 3, w, h, GL_RGB,
                              GL_UNSIGNED_BYTE, data);
        }
    }

    fprintf(stderr,"%d textures genereated\n",texnr+1);

    globe_id = glGenLists(1);
    glNewList(globe_id, GL_COMPILE);

//    glBegin(GL_QUADS);

    texnr=-1;
    for(ith=0;ith<nth;ith+=dth){
        qthnr=(nth-ith>=dth)?dth:nth-ith;

        for(iph=0;iph<nph;iph+=dph){
            qphnr=(nph-iph>=dph)?dph:nph-iph;
            texnr++;

            glBindTexture(GL_TEXTURE_2D, texbind[texnr]);
            glBegin(GL_QUADS);

            fprintf(stderr,"qthnr=%d\n",qthnr);
            for(qth=0;qth<qthnr;qth++){
                th1 = M_PI/(double)nth*(double)(ith+qth);
                th2 = M_PI/(double)nth*(double)(ith+qth+1);
                fprintf(stderr,"tex=%d, th1 = %f\n",texnr,th1);

                for(qph=0;qph<qphnr;qph++){
                    ph1 = -M_PI+2.0*M_PI/(double)nph*(double)(iph+qph);
                    ph2 = -M_PI+2.0*M_PI/(double)nph*(double)(iph+qph+1);

                    tex_x1 = (double)qph/(double)qphnr;
                    tex_x2 = (double)(qph+1)/(double)qphnr;
                    tex_y1 = (double)qth/(double)qthnr;
                    tex_y2 = (double)(qth+1)/(double)qthnr;

                    /* draw quad */
                    glTexCoord2f( tex_x1, tex_y1 );
                    glNormal3f(   sin(th1)*cos(ph1),   sin(th1)*sin(ph1),   cos(th1) );
                    glVertex3f( r*sin(th1)*cos(ph1), r*sin(th1)*sin(ph1), r*cos(th1) );
                    glTexCoord2f( tex_x2, tex_y1 );
                    glNormal3f(   sin(th1)*cos(ph2),   sin(th1)*sin(ph2),   cos(th1) );
                    glVertex3f( r*sin(th1)*cos(ph2), r*sin(th1)*sin(ph2), r*cos(th1) );
                    glTexCoord2f( tex_x2, tex_y2 );
                    glNormal3f(   sin(th2)*cos(ph2),   sin(th2)*sin(ph2),   cos(th2) );
                    glVertex3f( r*sin(th2)*cos(ph2), r*sin(th2)*sin(ph2), r*cos(th2) );
                    glTexCoord2f( tex_x1, tex_y2 );
                    glNormal3f(   sin(th2)*cos(ph1),   sin(th2)*sin(ph1),   cos(th2) );
                    glVertex3f( r*sin(th2)*cos(ph1), r*sin(th2)*sin(ph1), r*cos(th2) );

                }
            }

            glEnd(); /* QUADS */

        }
    }

    glEndList();

    free(data);
    free(alldata);

    return(globe_id);
}


#define HBYTEBITSUM(x) ( (((x)&0x08)!=0)+(((x)&0x04)!=0)+(((x)&0x02)!=0)+(((x)&0x01)!=0) )
#define BYTEBITSUM(x)  HBYTEBITSUM(x)+HBYTEBITSUM(x>>4)
#define WORDBITSUM(x)  BYTEBITSUM(x)+BYTEBITSUM(x>>8)
#define DWORDBITSUM(x) WORDBITSUM(x)+WORDBITSUM(x>>16)


int exists( char * filename )
{
    int rval=0;
    int f;
    f=open(filename,O_RDONLY);
    if( f!=-1 ){ rval=1; close(f); }
    return rval;
}


int make_globe( double r, int tex_th, int tex_ph, int dx, int dy, char * filename, char * map_path )
/* returns id */
{
    int ith,iph,x,y,w,h,dth,dph,qthnr,qphnr,qth,qph;
    char * data;
    float th1,th2,ph1,ph2, tex_x1,tex_x2,tex_y1,tex_y2;
    int globe_id;
    int allw,allh,alld;
    int nth, nph;
    unsigned char * alldata;
    static unsigned int texbind[10000];
    static int texnr=0;
    double x_new,y_new, x_d,y_d, val;
    int ic, allw_new, allh_new;
    int restbitsum;
    double xscale, yscale;
    int x_i,y_i;
    char fullfilename[256];

    if(texnr!=0){
        glDeleteTextures(texnr,texbind);
    }

    strcpy(fullfilename,map_path);
    strcat(fullfilename,filename);
    allw=-1;  allh=-1;
    if      ( strcasecmp(&filename[strlen(filename)-4],".jpg")==0 ){
        if      ( exists(filename) ){
            read_jpeg( filename, &allw, &allh, &alldata );
        }else if( exists(fullfilename) ){
            read_jpeg( fullfilename, &allw, &allh, &alldata );
        }
#ifdef DEBUG
        fprintf(stderr,"loaded jpg-file w/h:%d/%d\n",allw,allh);
#endif
    }else if( strcasecmp(&filename[strlen(filename)-4],".png")==0 ){
        if      ( exists(filename) ){
            load_png( filename, &allw, &allh, &alld, (char **)&alldata );
        }else if( exists(fullfilename) ){
            load_png( fullfilename, &allw, &allh, &alld, (char **)&alldata );
        }
    }
    //    alld=3;
    allh_new=allh;
    allw_new=allw;
//    fprintf(stderr,"!!!!!!!!!!bitsum(%X)=%d\n",allh_new,DWORDBITSUM(allh_new));
    restbitsum=DWORDBITSUM((allh%dy)/tex_th);
    if( restbitsum!=1 && restbitsum!=0 ){  /* rest is not a power of two */
        int dummy=tex_th;
        allh_new = allh/dy*dy+dummy;
        while(allh_new<allh){
            allh_new+=dummy;
            dummy*=2;
        }
//        allw_new = 2*allh_new;
    }
    restbitsum=DWORDBITSUM((allw%dx)/tex_ph);
    if( restbitsum!=1 && restbitsum!=0 ){  /* rest is not a power of two */
        int dummy=tex_ph;
        allw_new = allw/dx*dx+dummy;
        while(allw_new<allw){
            allw_new+=dummy;
            dummy*=2;
        }
//        allw_new = 2*allh_new;
    }

#ifdef DEBUG
    fprintf(stderr,"rescaling w/h:%d/%d\n",allw_new,allh_new);
#endif

    data = (char *)malloc(3*dx*dy);

    dth=dy/tex_th;
    dph=dx/tex_ph;
    nth=allh_new/tex_th;
    nph=allw_new/tex_ph;

    xscale=(double)allw/(double)allw_new;
    yscale=(double)allh/(double)allh_new;

    texnr=-1;
    for(ith=0;ith<nth;ith+=dth){
        for(iph=0;iph<nph;iph+=dph){
            /* get new quad texture */
            w=dx; h=dy;
            if( (iph/dph)*dx+w > allw_new ) w = allw_new % dx;
            if( (ith/dth)*dy+h > allh_new ) h = allh_new % dy;
            for(y=0;y<h;y++){
                if(allh==allh_new){
                    for(x=0;x<w;x++){
                        data[(y*w+x)*3+0]=alldata[(((ith/dth)*dy+y)*allw+(iph/dph)*dx+x)*3+0];
                        data[(y*w+x)*3+1]=alldata[(((ith/dth)*dy+y)*allw+(iph/dph)*dx+x)*3+1];
                        data[(y*w+x)*3+2]=alldata[(((ith/dth)*dy+y)*allw+(iph/dph)*dx+x)*3+2];
                    }
                } else {
                    y_new=(double)((ith/dth)*dy+y)*yscale;
                    x_new=(double)((iph/dph)*dx+0)*xscale;
                    y_d=y_new-floor(y_new);
                    y_i=(int)y_new;
                    for(x=0;x<w;x++){
                        x_i=(int)x_new;
                        x_d=x_new-floor(x_new);
                        for(ic=0;ic<3;ic++){
                            val=0.0;
                            val+=(1.0-x_d)*(1.0-y_d)*(double)alldata[((y_i+0)*allw+(x_i+0))*3+ic];
                            val+=   x_d   *(1.0-y_d)*(double)alldata[((y_i+0)*allw+(x_i+1))*3+ic];
                            val+=(1.0-x_d)*   y_d   *(double)alldata[((y_i+1)*allw+(x_i+0))*3+ic];
                            val+=   x_d   *   y_d   *(double)alldata[((y_i+1)*allw+(x_i+1))*3+ic];
                            data[(y*w+x)*3+ic]=val;
                        }
                        x_new+=xscale;
                    }
                }
            }
            texnr++;
#ifdef DEBUG
            fprintf(stderr,"tex = %d,   w = %d,   h = %d\n",texnr,w,h);
#endif
            glGenTextures( 1 , &texbind[texnr] );
            glBindTexture(GL_TEXTURE_2D,texbind[texnr]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, options_tex_min_filter);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, options_tex_mag_filter);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
            gluBuild2DMipmaps(GL_TEXTURE_2D, 3, w, h, GL_RGB,
                              GL_UNSIGNED_BYTE, data);
        }
    }

#ifdef DEBUG
    fprintf(stderr,"%d textures genereated\n",texnr+1);
#endif

    globe_id = glGenLists(1);
    glNewList(globe_id, GL_COMPILE);

//    glBegin(GL_QUADS);

    texnr=-1;
    for(ith=0;ith<nth;ith+=dth){
        qthnr=(nth-ith>=dth)?dth:nth-ith;

        for(iph=0;iph<nph;iph+=dph){
            qphnr=(nph-iph>=dph)?dph:nph-iph;
            texnr++;

            glBindTexture(GL_TEXTURE_2D, texbind[texnr]);
            glBegin(GL_QUADS);

#ifdef DEBUG
            fprintf(stderr,"qthnr=%d\n",qthnr);
#endif
            for(qth=0;qth<qthnr;qth++){
                th1 = M_PI/(double)nth*(double)(ith+qth);
                th2 = M_PI/(double)nth*(double)(ith+qth+1);
#ifdef DEBUG
                fprintf(stderr,"tex=%d, th1 = %f\n",texnr,th1);
#endif

                for(qph=0;qph<qphnr;qph++){
                    ph1 = -M_PI+2.0*M_PI/(double)nph*(double)(iph+qph);
                    ph2 = -M_PI+2.0*M_PI/(double)nph*(double)(iph+qph+1);

                    tex_x1 = (double)qph/(double)qphnr;
                    tex_x2 = (double)(qph+1)/(double)qphnr;
                    tex_y1 = (double)qth/(double)qthnr;
                    tex_y2 = (double)(qth+1)/(double)qthnr;

                    /* draw quad */
                    glTexCoord2f( tex_x1, tex_y1 );
                    glNormal3f(   sin(th1)*cos(ph1),   sin(th1)*sin(ph1),   cos(th1) );
                    glVertex3f( r*sin(th1)*cos(ph1), r*sin(th1)*sin(ph1), r*cos(th1) );
                    glTexCoord2f( tex_x2, tex_y1 );
                    glNormal3f(   sin(th1)*cos(ph2),   sin(th1)*sin(ph2),   cos(th1) );
                    glVertex3f( r*sin(th1)*cos(ph2), r*sin(th1)*sin(ph2), r*cos(th1) );
                    glTexCoord2f( tex_x2, tex_y2 );
                    glNormal3f(   sin(th2)*cos(ph2),   sin(th2)*sin(ph2),   cos(th2) );
                    glVertex3f( r*sin(th2)*cos(ph2), r*sin(th2)*sin(ph2), r*cos(th2) );
                    glTexCoord2f( tex_x1, tex_y2 );
                    glNormal3f(   sin(th2)*cos(ph1),   sin(th2)*sin(ph1),   cos(th2) );
                    glVertex3f( r*sin(th2)*cos(ph1), r*sin(th2)*sin(ph1), r*cos(th2) );

                }
            }

            glEnd(); /* QUADS */

        }
    }

    glEndList();

    free(data);
    free(alldata);

    return(globe_id);
}


void draw_globe(char * mapfile, char * map_path)
{
    static int init=0;
    static int globe_id=-1;

    if( !init || strcmp(globe_map_file,mapfile)!=0 ){
//        globe_id = make_globe_old2( 0.25, 24, 48, 32, 32, "photo_1536.png" );
//        globe_id = make_globe_old2( 0.25, 24, 48, 256, 256, "mggd_1536.png" );
//        globe_id = make_globe_old2( 0.25, 24, 48, 512, 512, "depths_1536.png" );
//        globe_id = make_globe( 0.25, 32, 32, 256, 256, "depths_1536.png" );
//        globe_id = make_globe( 0.25, 32, 32, 256, 256, "photo_1600.jpg" );
        if( mapfile==(char *)NULL ){
            strcpy( globe_map_file,"depths_1536.png" );
        } else {
            strcpy( globe_map_file, mapfile );
        }

        if(globe_id!=-1) glDeleteLists(globe_id,1);
        globe_id = make_globe( 0.25, 32, 32, 256, 256, globe_map_file, map_path );
//        globe_id = make_globe( 0.25, 32, 32, 256, 256, "depths_1536.png" );

//        globe_id = make_globe( 0.25, 32, 32, 256, 256, "depths_1440.jpg" );
        init=1;
    }

//    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glMaterialfv(GL_FRONT, GL_DIFFUSE,   col_diff);
    glMaterialfv(GL_FRONT, GL_AMBIENT,   col_amb );
    glMaterialfv(GL_FRONT, GL_SPECULAR,  col_spec);
    glMaterialf (GL_FRONT, GL_SHININESS, 0.0     );
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glCallList(globe_id);
}

void draw_dayline(VMvect sunpos, unsigned color )
{
    int i;
    VMvect n, nact, r;
    double r_grid;

    r_grid=0.25*GRID_R_FACT;

    sunpos = vec_unit(sunpos);

    r=vec_xyz(0.0,0.0,1.0);

    n = vec_unit(vec_diff(r,vec_proj(r,sunpos)));
    n = vec_scale(n,r_grid);
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE_MINUS_SRC_ALPHA,GL_SRC_ALPHA);
//    glColor4f(red,green,blue,alpha);
    glColor4ubv((GLubyte *)&color);
//    glColor4ub((GLubyte)(color>>16),(GLubyte)(color>>8),(GLubyte)color,(GLubyte)(color>>24));
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glBegin(GL_LINE_STRIP);
    for(i=0;i<100;i++){
        nact=n;
        rot_ax(sunpos,&nact,1,2.0*M_PI/100.0*(double)i);
        glVertex3f(nact.x,nact.y,nact.z);
    }
    glVertex3f(n.x,n.y,n.z);
    glEnd();
    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
}


void draw_grid( int nph, int nth, unsigned color )
{
    VMvect r;
    double th, dth, ph, dph, ddth,ddph, red,green,blue,alpha, r_grid;
    static int init=0;
    static int grid_id;

    r_grid=0.25*GRID_R_FACT;

    dph=2.0*M_PI/(double)nph;
    dth=M_PI/(double)nth;
    ddph=2.0*M_PI/(double)100;
    ddth=M_PI/(double)50;

    blue  = (double)( color      & 0xFF) / 255.0;
    green = (double)((color>>8)  & 0xFF) / 255.0;
    red   = (double)((color>>16) & 0xFF) / 255.0;
    alpha = 1.0-(double)((color>>24) & 0xFF) / 255.0;

    if (!init){
        init=1;
        grid_id = glGenLists(1);
        glNewList(grid_id, GL_COMPILE);
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE_MINUS_SRC_ALPHA,GL_SRC_ALPHA);
//        glColor4f(red,green,blue,alpha);
        glColor4ubv((GLubyte *)&color);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        /* longitudes */
        for(ph=0.0;ph<2.0*M_PI-dph/2.0;ph+=dph){
            glBegin(GL_LINE_STRIP);
            for(th=ddth;th<M_PI;th+=ddth){
                r.x = r_grid*sin(th)*cos(ph);
                r.y = r_grid*sin(th)*sin(ph);
                r.z = r_grid*cos(th);
                glVertex3f(r.x,r.y,r.z);
            }
            glEnd();
        }
        /* latitudes */
        for(th=0.0;th<M_PI;th+=dth){
            glBegin(GL_LINE_STRIP);
            for(ph=0.0;ph<2.0*M_PI;ph+=ddph){
                r.x = r_grid*sin(th)*cos(ph);
                r.y = r_grid*sin(th)*sin(ph);
                r.z = r_grid*cos(th);
                glVertex3f(r.x,r.y,r.z);
            }
            glEnd();
        }
        /* n polarkreis */
        th=23.4431*M_PI/180.0;
        glBegin(GL_LINES);
        for(ph=0.0;ph<2.0*M_PI;ph+=ddph){
            r.x = r_grid*sin(th)*cos(ph);
            r.y = r_grid*sin(th)*sin(ph);
            r.z = r_grid*cos(th);
            glVertex3f(r.x,r.y,r.z);
        }
        glEnd();
        /* s polarkreis */
        th=(180-23.4431)*M_PI/180.0;
        glBegin(GL_LINES);
        for(ph=0.0;ph<2.0*M_PI;ph+=ddph){
            r.x = r_grid*sin(th)*cos(ph);
            r.y = r_grid*sin(th)*sin(ph);
            r.z = r_grid*cos(th);
            glVertex3f(r.x,r.y,r.z);
        }
        glEnd();
        /* n wendekreis */
        th=(90.0-23.4431)*M_PI/180.0;;
        glBegin(GL_LINES);
        for(ph=0.0;ph<2.0*M_PI;ph+=ddph){
            r.x = r_grid*sin(th)*cos(ph);
            r.y = r_grid*sin(th)*sin(ph);
            r.z = r_grid*cos(th);
            glVertex3f(r.x,r.y,r.z);
        }
        glEnd();
        /* s wendekreis */
        th=(90.0+23.4431)*M_PI/180.0;;
        glBegin(GL_LINES);
        for(ph=0.0;ph<2.0*M_PI;ph+=ddph){
            r.x = r_grid*sin(th)*cos(ph);
            r.y = r_grid*sin(th)*sin(ph);
            r.z = r_grid*cos(th);
            glVertex3f(r.x,r.y,r.z);
        }
        glEnd();
        glEnable(GL_LIGHTING);
        glEnable(GL_TEXTURE_2D);
        glDisable(GL_BLEND);
        glEndList();
    }

    glCallList(grid_id);
}


void draw_grid_marks( int nph, int nth, unsigned color )
{
    VMvect r;
    double th, dth, ph, dph, ddth,ddph, red,green,blue,alpha, r_grid;
    static int init=0;
    static int grid_id;

    r_grid=0.25*GRID_R_FACT;

    dph=2.0*M_PI/(double)nph;
    dth=M_PI/(double)nth;
    ddph=2.0*M_PI/(double)100;
    ddth=M_PI/(double)50;

    blue  = (double)( color      & 0xFF) / 255.0;
    green = (double)((color>>8)  & 0xFF) / 255.0;
    red   = (double)((color>>16) & 0xFF) / 255.0;
    alpha = 1.0-(double)((color>>24) & 0xFF) / 255.0;

    if (!init){
        init=1;
        grid_id = glGenLists(1);
        glNewList(grid_id, GL_COMPILE);
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE_MINUS_SRC_ALPHA,GL_SRC_ALPHA);
//        glBlendFunc(GL_SRC_COLOR,GL_DST_COLOR);
//        glColor4f(red,green,blue,alpha);
        glColor4ubv((GLubyte *)&color);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        for(th=dth;th<M_PI;th+=dth){
        for(ph=0.0;ph<2.0*M_PI-dph/2.0;ph+=dph){
            glBegin(GL_LINES);
            /* longitudes */
              r.x = r_grid*sin(th)*cos(ph-0.03);
              r.y = r_grid*sin(th)*sin(ph-0.03);
              r.z = r_grid*cos(th);
              glVertex3f(r.x,r.y,r.z);
              r.x = r_grid*sin(th)*cos(ph+0.03);
              r.y = r_grid*sin(th)*sin(ph+0.03);
              r.z = r_grid*cos(th);
              glVertex3f(r.x,r.y,r.z);
              /* latitudes */
              r.x = r_grid*sin(th-0.03)*cos(ph);
              r.y = r_grid*sin(th-0.03)*sin(ph);
              r.z = r_grid*cos(th-0.03);
              glVertex3f(r.x,r.y,r.z);
              r.x = r_grid*sin(th+0.03)*cos(ph);
              r.y = r_grid*sin(th+0.03)*sin(ph);
              r.z = r_grid*cos(th+0.03);
              glVertex3f(r.x,r.y,r.z);
            glEnd();
        }}
        /* n polarkreis */
        th=23.4431*M_PI/180.0;
        glBegin(GL_LINES);
        for(ph=0.0;ph<2.0*M_PI;ph+=ddph){
            r.x = r_grid*sin(th)*cos(ph);
            r.y = r_grid*sin(th)*sin(ph);
            r.z = r_grid*cos(th);
            glVertex3f(r.x,r.y,r.z);
        }
        glEnd();
        /* s polarkreis */
        th=(180-23.4431)*M_PI/180.0;
        glBegin(GL_LINES);
        for(ph=0.0;ph<2.0*M_PI;ph+=ddph){
            r.x = r_grid*sin(th)*cos(ph);
            r.y = r_grid*sin(th)*sin(ph);
            r.z = r_grid*cos(th);
            glVertex3f(r.x,r.y,r.z);
        }
        glEnd();
        /* n wendekreis */
        th=(90.0-23.4431)*M_PI/180.0;;
        glBegin(GL_LINES);
        for(ph=0.0;ph<2.0*M_PI;ph+=ddph){
            r.x = r_grid*sin(th)*cos(ph);
            r.y = r_grid*sin(th)*sin(ph);
            r.z = r_grid*cos(th);
            glVertex3f(r.x,r.y,r.z);
        }
        glEnd();
        /* s wendekreis */
        th=(90.0+23.4431)*M_PI/180.0;;
        glBegin(GL_LINES);
        for(ph=0.0;ph<2.0*M_PI;ph+=ddph){
            r.x = r_grid*sin(th)*cos(ph);
            r.y = r_grid*sin(th)*sin(ph);
            r.z = r_grid*cos(th);
            glVertex3f(r.x,r.y,r.z);
        }
        glEnd();
        glEnable(GL_LIGHTING);
        glEnable(GL_TEXTURE_2D);
        glDisable(GL_BLEND);
        glEndList();
    }

    glCallList(grid_id);
}


void myGlWrite( VMvect pos, VMvect bx, VMvect by, int rel, double h, int bitmap, char *str )
{
    int i;
    static float base[]={
        0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0
    };
    static float m[16];
    float vx,vy,vz;

    VMvect bz;

//    fprintf(stderr,"drawing names \n");

    if( !bitmap ){

        if( rel ){
            glGetFloatv( GL_MODELVIEW_MATRIX, m );
            vx=bx.x; vy=bx.y; vz=bx.z;
            bx.x=vx*m[0]+vy*m[1]+vz*m[2];
            bx.y=vx*m[4]+vy*m[5]+vz*m[6];
            bx.z=vx*m[8]+vy*m[9]+vz*m[10];
            vx=by.x; vy=by.y; vz=by.z;
            by.x=vx*m[0]+vy*m[1]+vz*m[2];
            by.y=vx*m[4]+vy*m[5]+vz*m[6];
            by.z=vx*m[8]+vy*m[9]+vz*m[10];
        }

        bz=vec_cross(bx,by);
        bx=vec_scale(vec_unit(bx),h/119.05);
        by=vec_scale(vec_unit(by),h/119.05);
        bz=vec_scale(vec_unit(bz),h/119.05);

        base[ 0]=bx.x;  base[ 1]=bx.y;  base[ 2]=bx.z;
        base[ 4]=by.x;  base[ 5]=by.y;  base[ 6]=by.z;
        base[ 8]=bz.x;  base[ 9]=bz.y;  base[10]=bz.z;
        base[12]=pos.x; base[13]=pos.y; base[14]=pos.z;
        base[15]=1.0;

        glPushMatrix();
        glMultMatrixf(base);

         for(i=0;str[i]!=0;i++){
            glutStrokeCharacter(GLUT_STROKE_ROMAN,str[i]);
            glTranslatef(20.0,0.0,0.0);
        }

         glPopMatrix();

    } else {

        glRasterPos3f(pos.x,pos.y,pos.z);
        for(i=0;str[i]!=0;i++){
            glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12,str[i]);
//            glutBitmapCharacter(GLUT_BITMAP_8_BY_13,str[i]);
        }

    }
}


void draw_markers(char * markerfile, int bitmap_mode, unsigned mcol, unsigned lcol)
{
    static Marker * marker;
    static int init=0;
    static int markernr;
    int i;
    double th,ph;
    static int markers_id;
//    static float matrix[16];
    VMvect bx, by;
    float marker_h = 0.02;
    float m[16];
    float vx,vy,vz;

    if(!init){
        init=1;
        if(markerfile!=(char *)NULL){
            marker=load_markers(markerfile,&markernr);
        } else {
            marker=load_markers("xglobe-markers",&markernr);
        }
//        marker=load_markers("earth-markers-schaumann",&markernr);
#ifdef DEBUG
        fprintf(stderr,"markernr=%d\n",markernr);
        fprintf(stderr,"marker[0].name=%s\n",marker[0].name);
        fprintf(stderr,"marker[N-1].name=%s\n",marker[markernr-1].name);
#endif
        markers_id = glGenLists(1);
        glNewList(markers_id, GL_COMPILE);
//        glColor3f((double)((mcol>>16) & 0xFF)/255.0,
//                  (double)((mcol>> 8) & 0xFF)/255.0,
//                  (double)((mcol>> 0) & 0xFF)/255.0);
        glColor4ubv((GLubyte *)&mcol);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glBegin(GL_LINES);
        for(i=0;i<markernr;i++){
            th=(90.0-marker[i].lon)*M_PI/180.0;
            ph=marker[i].lat*M_PI/180.0;
            glVertex3f(0.25*sin(th)*cos(ph),
                       0.25*sin(th)*sin(ph),
                       0.25*cos(th));
            glVertex3f((0.25+marker_h)*sin(th)*cos(ph),
                       (0.25+marker_h)*sin(th)*sin(ph),
                       (0.25+marker_h)*cos(th));
        }
        glEnd();
        glEnable(GL_LIGHTING);
        glEnable(GL_TEXTURE_2D);
        glEndList();
    }

    glCallList(markers_id);

//    glColor3f((double)((lcol>>16) & 0xFF)/255.0,
//              (double)((lcol>> 8) & 0xFF)/255.0,
//              (double)((lcol>> 0) & 0xFF)/255.0);
    glColor4ubv((GLubyte *)&lcol);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);

    if(!bitmap_mode){
        bx=vec_xyz(1.0,0.0,0.0);
        by=vec_xyz(0.15,1.0,0.0);
        glGetFloatv( GL_MODELVIEW_MATRIX, m );
        vx=bx.x; vy=bx.y; vz=bx.z;
        bx.x=vx*m[0]+vy*m[1]+vz*m[2];
        bx.y=vx*m[4]+vy*m[5]+vz*m[6];
        bx.z=vx*m[8]+vy*m[9]+vz*m[10];
        vx=by.x; vy=by.y; vz=by.z;
        by.x=vx*m[0]+vy*m[1]+vz*m[2];
        by.y=vx*m[4]+vy*m[5]+vz*m[6];
        by.z=vx*m[8]+vy*m[9]+vz*m[10];
    }
    for(i=0;i<markernr;i++){
        th=(90.0-marker[i].lon)*M_PI/180.0;
        ph=marker[i].lat*M_PI/180.0;
        myGlWrite(
                  vec_xyz((0.25+marker_h*1.1)*sin(th)*cos(ph),
                          (0.25+marker_h*1.1)*sin(th)*sin(ph),
                          (0.25+marker_h*1.1)*cos(th)),
                  bx, by,
                  0,  /* 1/0: take screen/absolute directions for bx, by */
                  0.005,
                  bitmap_mode,  /* bitmap mode */
                  marker[i].name
                 );
    }
    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
}
